"""Decoding support for Aidon meters."""
from datetime import datetime
from typing import Dict, Union

import construct  # type: ignore

from amshan import cosem, obis_map

Element: construct.Struct = construct.Struct(
    construct.Const(
        cosem.CommonDataTypes.structure, cosem.CommonDataTypes
    ),  # expect structure
    "length" / construct.Int8ub,
    "obis" / cosem.ObisCodeOctedStringField,
    "content_type" / cosem.CommonDataTypes,
    "content"
    / construct.Switch(
        construct.this.content_type,
        {
            cosem.CommonDataTypes.visible_string: cosem.VisibleString,
            cosem.CommonDataTypes.octet_string: cosem.DateTime,
        },
        default=construct.Struct(
            "unscaled_value"
            / construct.Switch(
                construct.this._.content_type,
                {
                    cosem.CommonDataTypes.double_long_unsigned: cosem.DoubleLongUnsigned,
                    cosem.CommonDataTypes.long: cosem.Long,
                    cosem.CommonDataTypes.long_unsigned: cosem.LongUnsigned,
                },
            ),
            "scaler_unit" / cosem.ScalerUnitField,
            "value"
            / construct.Computed(
                construct.this.unscaled_value * construct.this.scaler_unit.scaler.scale
            ),
        ),
    ),
)

NotificationBody: construct.Struct = construct.Struct(
    construct.Const(cosem.CommonDataTypes.array, cosem.CommonDataTypes),  # expect array
    "length" / construct.Int8ub,
    "list_items" / construct.Array(construct.this.length, Element),
)

LlcPdu: construct.Struct = cosem.get_llc_pdu_struct(NotificationBody)


def normalize_parsed_frame(
    frame: construct.Struct,
) -> Dict[str, Union[str, int, float, datetime]]:
    """Convert data from meters construct structure to a dictionary with common key names."""
    list_items = frame.information.notification_body.list_items

    dictionary: Dict[str, Union[str, int, float, datetime]] = {
        obis_map.NEK_HAN_FIELD_METER_MANUFACTURER: "Aidon"
    }
    for measure in list_items:
        element_name = obis_map.obis_name_map[measure.obis]

        if isinstance(measure.content, str):
            dictionary[element_name] = measure.content
        else:
            if hasattr(measure.content, "datetime"):
                dictionary[element_name] = measure.content.datetime
            else:
                dictionary[element_name] = (
                    measure.content.unscaled_value
                    if measure.content.unscaled_value == measure.content.value
                    else float(measure.content.value)
                )

    return dictionary


def decode_frame_content(
    frame_content: bytes,
) -> Dict[str, Union[str, int, float, datetime]]:
    """Decode meter LLC PDU frame content as a dictionary."""
    parsed = LlcPdu.parse(frame_content)
    return normalize_parsed_frame(parsed)



"""Mappings between OBIS codes and keys used in decoded data."""
from typing import Dict, List

NEK_HAN_FIELD_OBIS_LIST_VER_ID = "list_ver_id"
NEK_HAN_FIELD_METER_ID = "meter_id"
NEK_HAN_FIELD_METER_TYPE = "meter_type"
NEK_HAN_FIELD_METER_MANUFACTURER = "meter_manufacturer"
NEK_HAN_FIELD_METER_DATETIME = "meter_datetime"
NEK_HAN_FIELD_ACTIVE_POWER_IMPORT = "active_power_import"
NEK_HAN_FIELD_ACTIVE_POWER_EXPORT = "active_power_export"
NEK_HAN_FIELD_REACTIVE_POWER_IMPORT = "reactive_power_import"
NEK_HAN_FIELD_REACTIVE_POWER_EXPORT = "reactive_power_export"
NEK_HAN_FIELD_CURRENT_L1 = "current_l1"
NEK_HAN_FIELD_CURRENT_L2 = "current_l2"
NEK_HAN_FIELD_CURRENT_L3 = "current_l3"
NEK_HAN_FIELD_VOLTAGE_L1 = "voltage_l1"
NEK_HAN_FIELD_VOLTAGE_L2 = "voltage_l2"
NEK_HAN_FIELD_VOLTAGE_L3 = "voltage_l3"
NEK_HAN_FIELD_ACTIVE_POWER_IMPORT_HOUR = "active_power_import_hour"
NEK_HAN_FIELD_ACTIVE_POWER_EXPORT_HOUR = "active_power_export_hour"
NEK_HAN_FIELD_REACTIVE_POWER_IMPORT_HOUR = "reactive_power_import_hour"
NEK_HAN_FIELD_REACTIVE_POWER_EXPORT_HOUR = "reactive_power_export_hour"

name_obis_map: Dict[str, List[str]] = {
    NEK_HAN_FIELD_OBIS_LIST_VER_ID: ["1.1.0.2.129.255"],
    NEK_HAN_FIELD_METER_ID: ["0.0.96.1.0.255", "1.1.0.0.5.255"],
    NEK_HAN_FIELD_METER_TYPE: ["0.0.96.1.7.255", "1.1.96.1.1.255"],
    NEK_HAN_FIELD_ACTIVE_POWER_IMPORT: ["1.0.1.7.0.255", "1.1.1.7.0.255"],
    NEK_HAN_FIELD_ACTIVE_POWER_EXPORT: ["1.0.2.7.0.255", "1.1.2.7.0.255"],
    NEK_HAN_FIELD_REACTIVE_POWER_IMPORT: ["1.0.3.7.0.255", "1.1.3.7.0.255"],
    NEK_HAN_FIELD_REACTIVE_POWER_EXPORT: ["1.0.4.7.0.255", "1.1.4.7.0.255"],
    NEK_HAN_FIELD_CURRENT_L1: ["1.0.31.7.0.255", "1.1.31.7.0.255"],
    NEK_HAN_FIELD_CURRENT_L2: ["1.0.51.7.0.255", "1.1.51.7.0.255"],
    NEK_HAN_FIELD_CURRENT_L3: ["1.0.71.7.0.255", "1.1.71.7.0.255"],
    NEK_HAN_FIELD_VOLTAGE_L1: ["1.0.32.7.0.255", "1.1.32.7.0.255"],
    NEK_HAN_FIELD_VOLTAGE_L2: ["1.0.52.7.0.255", "1.1.52.7.0.255"],
    NEK_HAN_FIELD_VOLTAGE_L3: ["1.0.72.7.0.255", "1.1.72.7.0.255"],
    NEK_HAN_FIELD_METER_DATETIME: ["0.0.1.0.0.255", "0.1.1.0.0.255"],
    NEK_HAN_FIELD_ACTIVE_POWER_IMPORT_HOUR: ["1.0.1.8.0.255", "1.1.1.8.0.255"],
    NEK_HAN_FIELD_ACTIVE_POWER_EXPORT_HOUR: ["1.0.2.8.0.255", "1.1.2.8.0.255"],
    NEK_HAN_FIELD_REACTIVE_POWER_IMPORT_HOUR: ["1.0.3.8.0.255", "1.1.3.8.0.255"],
    NEK_HAN_FIELD_REACTIVE_POWER_EXPORT_HOUR: ["1.0.4.8.0.255", "1.1.4.8.0.255"],
}

obis_name_map: Dict[str, str] = {}
for name, obis_values in name_obis_map.items():
    for obis in obis_values:
        obis_name_map[obis] = name



"""Use this module to easily decode any supported meter frame format."""
from datetime import datetime
from typing import Dict, Optional, Union

import construct  # type: ignore

from amshan import aidon, kaifa, kamstrup


class AutoDecoder:
    """
    Use this class to easily decode any supported meter frame formats into a common dictionary format.

    The class tries all decoders until success or no more decoders.
    The successful decoder is stored and tried as the first decoder next time.
    """

    decoder_functions = [
        ("Aidon", aidon.decode_frame_content),
        ("Kaifa", kaifa.decode_frame_content),
        ("Kamstrup", kamstrup.decode_frame_content),
    ]

    def __init__(self) -> None:
        """Initialize AutoDecoder."""
        self.__previous_success: Optional[int] = None

    @property
    def previous_success_decoder(self) -> Optional[str]:
        """Return the name of the previous successful decoder function, or None if none previous success."""
        if self.__previous_success is not None:
            decoder_name, _ = AutoDecoder.decoder_functions[self.__previous_success]
            return decoder_name
        return None

    def decode_frame_content(
        self, frame_content: bytes
    ) -> Optional[Dict[str, Union[str, int, float, datetime]]]:
        """
        Decode meter LLC PDU frame content as a dictionary.

        The previous meter decoder used with success is used first. Then other meter decoders are tried.
        :rtype: dictionary or None if none of the decoders worked.
        """
        previous_success_index = (
            self.__previous_success if self.__previous_success else 0
        )

        for i in range(len(AutoDecoder.decoder_functions)):
            index = (i + previous_success_index) % len(AutoDecoder.decoder_functions)
            _, decoder = AutoDecoder.decoder_functions[index]
            try:
                decoded_frame = decoder(frame_content)
                self.__previous_success = index
                return decoded_frame
            except construct.ConstructError:
                pass

        return None



"""Contruct declarations for some COSEM types and structures."""
import datetime
from decimal import Decimal
from typing import Any

import construct  # type: ignore

# See COSEM blue Book table 2 (Common data types) in section 4.1.5 Common data types
# NOTE: Not all types are defined here
CommonDataTypes = construct.Enum(
    construct.Int8ub,
    null_data=0,
    array=1,
    structure=2,
    double_long_unsigned=6,  # Unsigned32 [0...4 294 967 295]
    octet_string=9,  # An ordered sequence of octets (8 bit bytes)
    visible_string=10,  # An ordered sequence of ASCII characters
    integer=15,  # Integer8 [-128...127]
    long=16,  # Integer16 [-32 768...32 767]
    long_unsigned=18,  # Unsigned16 [0...65 535]
    enum=22,  # as CosemPhysicalUnits
)

# See COSEM blue Book table 4 (Enumerated values for physical units) in section 4.3.2 Register
# NOTE: Not all values are defined here
PhysicalUnits = construct.Enum(
    construct.Int8ub, W=27, var=29, Wh=30, varh=32, A=33, V=35
)

# Integer8 [-128...127]
Integer = construct.Int8sb

# Integer16 [-32 768...32 767]
Long = construct.Int16sb

# Unsigned16 [0...65 535]
LongUnsigned = construct.Int16ub

# Unsigned32 [0...4 294 967 295]
DoubleLongUnsigned = construct.Int32ub

VisibleString = construct.PascalString(construct.Int8ub, "ASCII")

OctedStringText = construct.FocusedSeq(
    "value",
    "length" / construct.Int8ub,
    "value" / construct.PaddedString(construct.this.length, "ASCII"),
)

ObisCode = construct.ExprAdapter(
    construct.Int8ub[6],
    decoder=lambda obj, ctx: ".".join(f"{b}" for b in obj),
    encoder=lambda obj, ctx: [int(part) for part in obj.split(".")],
)


def _type_code_to_type(
    type_code: construct.Enum,
) -> Any:
    return construct.Switch(
        type_code,
        {
            CommonDataTypes.integer: Integer,
            CommonDataTypes.long: Long,
            CommonDataTypes.long_unsigned: LongUnsigned,
            CommonDataTypes.double_long_unsigned: DoubleLongUnsigned,
            CommonDataTypes.visible_string: VisibleString,
        },
        default=construct.Error,
    )


OptionalDateTimeByte = construct.ExprAdapter(
    construct.Int8ub,
    decoder=lambda obj, ctx: obj if obj != 0xFF else None,
    encoder=lambda obj, ctx: obj if obj is not None else 0xFF,
)

# See COSEM blue Book section 4.1.6.1 Date and time formats
DateTime = construct.Struct(
    construct.Const(0x0C, construct.Int8ub),  # expect length 12
    "year" / construct.Int16ub,
    "month" / construct.Int8ub,
    "day_of_month" / construct.Int8ub,
    "day_of_week" / construct.Int8ub,
    "hour" / OptionalDateTimeByte,
    "minute" / OptionalDateTimeByte,
    "second" / OptionalDateTimeByte,
    "hundredths_of_second" / OptionalDateTimeByte,
    "deviation"
    / construct.ExprAdapter(
        construct.Int16sb,
        decoder=lambda obj, ctx: obj if obj != -0x8000 else None,
        encoder=lambda obj, ctx: obj if obj is not None else -0x8000,
    ),
    "clock_status_byte" / construct.Peek(OptionalDateTimeByte),
    "clock_status"
    / construct.If(
        construct.this.clock_status_byte != 0xFF,
        construct.BitStruct(
            "invalid_value" / construct.BitsInteger(1),
            "doubtful_value" / construct.BitsInteger(1),
            "different_clock_base" / construct.BitsInteger(1),
            "invalid_clock_status" / construct.BitsInteger(1),
            construct.BitsInteger(3),
            "daylight_saving_active" / construct.BitsInteger(1),
        ),
    ),
    construct.If(construct.this.clock_status_byte == 0xFF, construct.Int8ub),
    "datetime"
    / construct.Computed(
        lambda ctx: datetime.datetime(
            ctx.year,
            ctx.month,
            ctx.day_of_month,
            ctx.hour,
            ctx.minute,
            ctx.second,
            ctx.hundredths_of_second * 10000
            if ctx.hundredths_of_second is not None
            else 0,
            datetime.timezone(datetime.timedelta(minutes=ctx.deviation))
            if ctx.deviation is not None
            else None,
        )
    ),
)

# field types

Field = construct.FocusedSeq(
    "value",
    "content_type" / CommonDataTypes,
    "value" / _type_code_to_type(construct.this.content_type),
)

ObisCodeOctedStringField = construct.FocusedSeq(
    "code",
    construct.Const(
        CommonDataTypes.octet_string, CommonDataTypes
    ),  # expect octet string (bytes)
    construct.Const(6, construct.Int8ub),  # expect length 6
    "code" / ObisCode,
)

VisibleStringField = construct.FocusedSeq(
    "value",
    construct.Const(
        CommonDataTypes.visible_string, CommonDataTypes
    ),  # An ordered sequence of ASCII characters
    "value" / VisibleString,
)

OctedStringTextField = construct.FocusedSeq(
    "value",
    construct.Const(
        CommonDataTypes.octet_string, CommonDataTypes
    ),  # expect octet string (bytes)
    "value" / OctedStringText,
)

IntegerField = construct.FocusedSeq(
    "value",
    construct.Const(CommonDataTypes.integer, CommonDataTypes),  # expect integer
    "value" / Integer,
)

DoubleLongUnsignedField = construct.FocusedSeq(
    "value",
    construct.Const(CommonDataTypes.double_long_unsigned, CommonDataTypes),
    "value" / DoubleLongUnsigned,
)

UnitField = construct.FocusedSeq(
    "value",
    construct.Const(CommonDataTypes.enum, CommonDataTypes),  # expect enum
    "value" / PhysicalUnits,
)

Scaler = construct.Struct(
    "exponent"
    / IntegerField,  # This is the exponent (to the base of 10) of the multiplication factor.
    "scale" / construct.Computed(lambda ctx: Decimal(10) ** ctx.exponent),
)

ScalerUnitField = construct.Struct(
    construct.Const(CommonDataTypes.structure, CommonDataTypes),  # expect structure
    construct.Const(2, construct.Int8ub),  # expect length 2
    "scaler" / Scaler,
    "unit" / UnitField,
)

DateTimeField = construct.FocusedSeq(
    "value",
    construct.Const(CommonDataTypes.octet_string, CommonDataTypes),
    "value" / DateTime,
)

OptionalDateTimeField = construct.FocusedSeq(
    "value",
    "content_type" / CommonDataTypes,
    construct.Check(
        lambda ctx: ctx.content_type
        in (CommonDataTypes.null_data, CommonDataTypes.octet_string)
    ),
    "value"
    / construct.If(construct.this.content_type != CommonDataTypes.null_data, DateTime),
)

LongInvokeIdAndPriority = construct.BitStruct(
    "invoke-id" / construct.BitsInteger(24),
    construct.Padding(4),
    "self-descriptive"
    / construct.Enum(construct.BitsInteger(1), NotSelfDescriptive=0, SelfDescriptive=1),
    "processing-option"
    / construct.Enum(construct.BitsInteger(1), ContinueOnError=0, BreakOnError=1),
    "service-class"
    / construct.Enum(construct.BitsInteger(1), Unconfirmed=0, Confirmed=1),
    "priority" / construct.Enum(construct.BitsInteger(1), Normal=0, High=1),
)


def _get_apdpu_struct(notification_body: construct.Struct) -> construct.Struct:
    return construct.Struct(
        "Tag" / construct.Int8ub,
        "LongInvokeIdAndPriority" / LongInvokeIdAndPriority,
        "_datetimestartbyte" / construct.Peek(CommonDataTypes),
        "DateTime"
        / construct.Switch(
            construct.this._datetimestartbyte,  # pylint: disable=protected-access
            {
                CommonDataTypes.null_data: construct.Byte,
                CommonDataTypes.octet_string: DateTimeField,
            },
            default=DateTime,
        ),
        "notification_body" / notification_body,
    )


def get_llc_pdu_struct(notification_body: construct.Struct) -> construct.Struct:
    """Get a LLC PDU struct wrapping supplied notification body."""
    return construct.Struct(
        "dsap" / construct.Int8ub,
        "ssap" / construct.Int8ub,
        "control" / construct.Int8ub,
        "information" / _get_apdpu_struct(notification_body),
    )



"""Async meter connection module."""
import datetime
import logging
from abc import ABCMeta, abstractmethod
from asyncio import (
    FIRST_COMPLETED,
    BaseTransport,
    CancelledError,
    Event,
    Future,
    Protocol,
    Queue,
    iscoroutinefunction,
    wait,
    sleep,
)
from typing import Awaitable, Callable, ClassVar, Optional, Tuple

from amshan import hdlc
from amshan.hdlc import HdlcFrame

_LOGGER = logging.getLogger(__name__)


class BackOffStrategy(metaclass=ABCMeta):
    """
    Back-off strategy base class.

    Create sub-classes to implement different strategies.
    """

    DEFAULT_MAX_DELAY_SEC: int = 60

    @abstractmethod
    def failure(self) -> None:
        """Call this after a failure."""

    @abstractmethod
    def reset(self) -> None:
        """Call this after success to reset."""

    @property
    @abstractmethod
    def current_delay_sec(self) -> int:
        """Return current back-off delay in seconds."""


class ExponentialBackOff(BackOffStrategy):
    """Exponential back-off strategy."""

    def __init__(self) -> None:
        """Initialize ExponentialBackOff."""
        self._delay: int = 0
        self.max_delay: int = super().DEFAULT_MAX_DELAY_SEC

    def failure(self) -> None:
        """Call this after a failure."""
        self._delay = self._delay * 2
        if self._delay == 0:
            self._delay = 1

    def reset(self) -> None:
        """Call this after success to reset."""
        self._delay = 0

    @property
    def current_delay_sec(self) -> int:
        """Return current back-off delay in seconds."""
        return self._delay if self._delay < self.max_delay else self.max_delay


class SmartMeterBaseProtocol(Protocol, metaclass=ABCMeta):
    """
    Network protocol base class that reads smart meter frames from a stream.

    Sub classes must implement frame_received().
    """

    # Total number of this class that has been created.
    # The number is used when initializing trace id for new instances.
    total_instance_counter: ClassVar[int] = 0

    def __init__(
        self,
        frame_reader: Optional[hdlc.HdlcFrameReader] = None,
    ) -> None:
        """
        Initialize SmartMeterProtocol.

        :param frame_reader: optional frame reader.
        A frame reader with both octet stuffing and abort sequence disabled is created if None.
        """
        super().__init__()
        self._done: Future[None] = Future()
        self.instance_id: int = SmartMeterBaseProtocol.total_instance_counter
        self._frame_reader = (
            frame_reader
            if frame_reader
            else hdlc.HdlcFrameReader(use_octet_stuffing=False, use_abort_sequence=True)
        )
        self._transport: Optional[BaseTransport] = None
        self._transport_info: Optional[str] = None
        SmartMeterBaseProtocol.total_instance_counter += 1

    @property
    def done(self) -> Awaitable[None]:
        """Return Awaitable that can be used to wait for connection to be lost or closed."""
        return self._done

    def _set_transport_info(self) -> None:
        if self._transport:
            if hasattr(self._transport, "serial"):
                self._transport_info = str(self._transport.serial)  # type: ignore
            else:
                peer_name = self._transport.get_extra_info("peername")
                if peer_name:
                    host, port, *_ = peer_name
                    self._transport_info = f"host {host} and port {port}"

            if not self._transport_info:
                self._transport_info = str(self._transport)

    def connection_made(self, transport: BaseTransport) -> None:
        """
        Connect the transport with this protocol instance.

        Called when a connection is made.
        The argument is the transport representing the connection.
        To receive data, wait for data_received() calls.
        When the connection is closed, connection_lost() is called.
        """
        self._transport = transport
        self._set_transport_info()
        _LOGGER.info(
            "%s: Smart meter connected to %s", self._instance_id(), self._transport_info
        )

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """
        Signal that the transport connected to this protocol instance has been lost or closed.

        Called when the connection is lost or closed.
        The argument is an exception object or None (the latter
        meaning a regular EOF is received or the connection was
        aborted or closed).

        This method set the Done property.
        """
        if exc:
            _LOGGER.warning(
                "%s: Connection to %s lost: %s",
                self._instance_id(),
                self._transport_info,
                exc,
            )
        else:
            _LOGGER.debug(
                "%s: Connection to %s closed",
                self._instance_id(),
                self._transport_info,
            )

        if self._transport:
            try:
                self._transport.close()
                self._transport = None
            except Exception as ex:  # pylint: disable=broad-except
                _LOGGER.warning(
                    "%s: Error when closing transport %s for %s connection: %s",
                    self._instance_id(),
                    self._transport_info,
                    "lost" if exc else "closed",
                    ex,
                )

        self._done.set_result(None)

    def data_received(self, data: bytes) -> None:
        """Receive data from the transport and put frame(s) on the queue if frame(s) are complete."""
        frames = self._frame_reader.read(data)
        for frame in frames:
            self.frame_received(frame)

    @abstractmethod
    def frame_received(self, frame: HdlcFrame) -> None:
        """Frame is received from the transport."""

    def eof_received(self) -> bool:
        """
        Return False to close the transport.

        Called when the other end signals it won’t send any more data.
        """
        _LOGGER.debug(
            "%s: eof_received - the other end (%s) signaled it won’t send any more data. Close transport",
            self._instance_id(),
            self._transport_info,
        )

        # return false to close transport.
        return False

    def _instance_id(self) -> str:
        return f"{self.__class__.__name__}[{self.instance_id}]"


# pylint: disable=too-many-instance-attributes


class SmartMeterFrameProtocol(SmartMeterBaseProtocol):
    """
    Network protocol that reads smart meter frames from a stream and forwards them to a queue.

    When the user wants to requests a transport to use with this protocol,
    they pass a SmartMeterFrameProtocol factory to a utility function (e.g.,
    EventLoop.create_connection() or serial_asyncio.create_serial_connection).

    Example:
        await serial_asyncio.create_serial_connection(loop, lambda: SmartMeterFrameProtocol(queue), url = "/dev/tty01")
    """

    def __init__(
        self,
        destination_queue: "Queue[hdlc.HdlcFrame]",
        frame_reader: Optional[hdlc.HdlcFrameReader] = None,
    ) -> None:
        """
        Initialize SmartMeterProtocol.

        :param destination_queue: destination queue for decoded frames.
        :param frame_reader: optional frame reader.
        A frame reader with both octet stuffing and abort sequence disabled is created if None.
        """
        super().__init__(frame_reader)
        self.queue: Queue[HdlcFrame] = destination_queue

    def frame_received(self, frame: HdlcFrame) -> None:
        """Frame is passed on to the queue."""
        self.queue.put_nowait(frame)


class SmartMeterFrameContentProtocol(SmartMeterBaseProtocol):
    """
    Network protocol that reads smart meter frames from a stream and forwards their content to a queue.

    When the user wants to requests a transport to use with this protocol,
    they pass a SmartMeterFrameContentProtocol factory to a utility function (e.g.,
    EventLoop.create_connection() or serial_asyncio.create_serial_connection).

    Example:
        await serial_asyncio.create_serial_connection(loop, lambda: SmartMeterFrameContentProtocol(queue), url = "/dev/tty01")
    """

    def __init__(
        self,
        destination_queue: "Queue[bytes]",
        frame_reader: Optional[hdlc.HdlcFrameReader] = None,
    ) -> None:
        """
        Initialize SmartMeterProtocol.

        :param destination_queue: destination queue for decoded frames.
        :param frame_reader: optional frame reader.
        A frame reader with both octet stuffing and abort sequence disabled is created if None.
        """
        super().__init__(frame_reader)
        self.queue: Queue[bytes] = destination_queue

    def frame_received(self, frame: HdlcFrame) -> None:
        """
        Frame is received and its content is passed on to the queue.

        Only content from non empty frames with expected length and checksum is passed on.
        """
        if frame.is_good_ffc and frame.is_expected_length:
            frame_content = frame.information
            if frame_content:
                self.queue.put_nowait(frame_content)
            else:
                _LOGGER.debug("Got empty frame")
        else:
            _LOGGER.warning("Got invalid frame: %s", frame.frame_data.hex())


MeterTransportProtocol = Tuple[BaseTransport, SmartMeterBaseProtocol]

AsyncConnectionFactory = Callable[[], Awaitable[MeterTransportProtocol]]


class ConnectionManager:
    """
    Maintain connection and reconnect if connection is lost.

    Reconnecting uses a back-off retry strategy, and has a simple circuit breaker for connection lost.
    """

    DEFAULT_CONNECTION_LOST_BACK_OFF_THRESHOLD: int = 5
    DEFAULT_CONNECTION_LOST_BACK_OFF_SLEEP_SEC: int = 5

    def __init__(
        self,
        connection_factory: AsyncConnectionFactory,
    ) -> None:
        """
        Initialize class.

        :param connection_factory: A factory function that returns a Transport and SmartMeterProtocol tuple.
        """
        if not iscoroutinefunction(connection_factory):
            raise ValueError("Factory must be awaitable.")

        self._connection_factory: AsyncConnectionFactory = connection_factory
        self._connection: Optional[MeterTransportProtocol] = None
        self._is_closing: Event = Event()

        self.back_off_connect_error: BackOffStrategy = ExponentialBackOff()

        self.connection_lost_back_off_threshold: int = (
            ConnectionManager.DEFAULT_CONNECTION_LOST_BACK_OFF_SLEEP_SEC
        )
        self.connection_lost_back_off_sleep_sec: int = (
            ConnectionManager.DEFAULT_CONNECTION_LOST_BACK_OFF_SLEEP_SEC
        )
        self._connection_lost_last_time: Optional[datetime.datetime] = None
        self._connection_lost_sleep_before_reconnect: bool = False

    def close(self) -> None:
        """Close current connection, if any, and stop reconnecting."""
        self._is_closing.set()
        if self._connection:
            _LOGGER.info("Close connection and abort connect loop")
            transport, _ = self._connection
            transport.close()
            self._connection = None

    async def connect_loop(self) -> None:
        """
        Connect to meter using connection factory, and keep reconnecting if connection is lost.

        The connection is not reconnected on connection loss if close() was called on this instance.
        """
        while not self._is_closing.is_set():
            await wait(
                (self._try_connect(), self._is_closing.wait()),
                return_when=FIRST_COMPLETED,
            )

            if self._connection:
                _, protocol = self._connection
                await wait(
                    (protocol.done, self._is_closing.wait()),
                    return_when=FIRST_COMPLETED,
                )

                if not self._is_closing.is_set():
                    _LOGGER.warning("Connection lost")
                    self._update_connection_lost_circuit_breaker()

                self._connection = None

        # done closing if that was the case of connection loss
        self._is_closing.clear()

        _LOGGER.info("Connect loop done")

    def _update_connection_lost_circuit_breaker(self) -> None:
        now = datetime.datetime.utcnow()
        if self._connection_lost_last_time:
            delta = now - self._connection_lost_last_time
            self._connection_lost_sleep_before_reconnect = (
                delta.total_seconds() < self.connection_lost_back_off_threshold
            )
        self._connection_lost_last_time = now

    def _get_back_off_time(self) -> int:
        sleep_time = 0
        current_connect_error_delay = self.back_off_connect_error.current_delay_sec
        if (
            current_connect_error_delay > 0
            or self._connection_lost_sleep_before_reconnect
        ):
            reconnect_sleep = (
                self.connection_lost_back_off_sleep_sec
                if self._connection_lost_sleep_before_reconnect
                else 0
            )
            sleep_time = max(current_connect_error_delay, reconnect_sleep)

            _LOGGER.info(
                "Back-off for %d sec before reconnecting",
                sleep_time,
            )
        return sleep_time

    async def _try_connect(self) -> None:
        sleep_time = self._get_back_off_time()
        if sleep_time > 0:
            await sleep(sleep_time)

        if not self._is_closing.is_set():
            try:
                _LOGGER.debug("Try to connect")
                self._connection = await self._connection_factory()

                self.back_off_connect_error.reset()
            except CancelledError:
                _LOGGER.debug("The operation was cencelled")
                raise
            except Exception as ex:  # pylint: disable=broad-except
                self._connection = None
                self.back_off_connect_error.failure()
                _LOGGER.warning("Error connecting: %s", ex)



"""TCP/IP connection factory."""
from asyncio import AbstractEventLoop, BaseProtocol, get_event_loop
from asyncio import Queue  # pylint: disable=unused-import
from typing import Optional, cast

from amshan.hdlc import HdlcFrame  # pylint: disable=unused-import
from amshan.meter_connection import (
    MeterTransportProtocol,
    SmartMeterFrameContentProtocol,
    SmartMeterFrameProtocol,
)


async def create_tcp_frame_connection(
    queue: "Queue[HdlcFrame]",
    loop: Optional[AbstractEventLoop],
    *args,
    **kwargs,
) -> MeterTransportProtocol:
    """
    Create TCP connection using SmartMeterFrameProtocol.

    :param queue: Queue for received frames
    :param loop: The event handler
    :param args: Passed to the loop.create_connection
    :param kwargs: Passed to the loop.create_connection
    :return: Tuple of transport and protocol
    """
    loop = loop if loop else get_event_loop()
    return cast(
        MeterTransportProtocol,
        await loop.create_connection(
            lambda: cast(BaseProtocol, SmartMeterFrameProtocol(queue)),
            *args,
            **kwargs,
        ),
    )


async def create_tcp_frame_content_connection(
    queue: "Queue[bytes]",
    loop: Optional[AbstractEventLoop],
    *args,
    **kwargs,
) -> MeterTransportProtocol:
    """
    Create TCP connection using SmartMeterFrameContentProtocol.

    :param queue: Queue for received frames
    :param loop: The event handler
    :param args: Passed to the loop.create_connection
    :param kwargs: Passed to the loop.create_connection
    :return: Tuple of transport and protocol
    """
    loop = loop if loop else get_event_loop()
    return cast(
        MeterTransportProtocol,
        await loop.create_connection(
            lambda: cast(BaseProtocol, SmartMeterFrameContentProtocol(queue)),
            *args,
            **kwargs,
        ),
    )



"""Compute or check checksum using a 16-bit Fast Frame Check Sequence (FCS) derived from RFC1662."""
from typing import List


def _compute_fcs_16_crc_table() -> List[int]:
    """Generate a FCS-16 table."""
    polynomial = 0x8408  # The FCS-16 generator polynomial: x**0 + x**5 + x**12 + x**16.
    crc_table = [] * 256
    for byte in range(256):
        crc = 0
        for _ in range(8):
            if (byte ^ crc) & 1:
                crc = (crc >> 1) ^ polynomial
            else:
                crc >>= 1
            byte >>= 1
        crc_table.append(crc)
    return crc_table


class FastFrameCheckSequence16:
    """16 bit Fast Frame Check Sequence (FCS)."""

    INIT_FCS_16 = 0xFFFF  # Initial FCS value
    GOOD_FCS_16 = 0xF0B8  # Good final FCS value

    fast_frame_check_crc_table = _compute_fcs_16_crc_table()

    def __init__(self) -> None:
        """Initialize FastFrameCheckSequence16."""
        self._crc_value = self.INIT_FCS_16

    @staticmethod
    def _next(crc: int, byte: int) -> int:
        """Calculate a new fcs CRC given the current CRC value and the new data."""
        crc_index = (crc ^ byte) & 0xFF
        return (crc >> 8) ^ FastFrameCheckSequence16.fast_frame_check_crc_table[
            crc_index
        ]

    def update(self, byte: int) -> int:
        """Update the calculated CRC value for the specified input data."""
        self._crc_value = self._next(self._crc_value, byte)
        return self._crc_value

    @property
    def is_good(self) -> bool:
        """
        Return True when data AND checksum has been processed with update(), and the checksum is correct.

        FCS was designed so that a particular pattern results
        when the FCS operation passes over the complemented FCS.  A good
        frame is indicated by this "good FCS" value
        """
        return self.GOOD_FCS_16 == self._crc_value

    @property
    def checksum(self) -> int:
        """Return current checksum."""
        return self._crc_value ^ 0xFFFF  # complement

    @staticmethod
    def compute_checksum(data: bytes, start: int, length: int) -> int:
        """Compute checksum of bytes."""
        fcs = FastFrameCheckSequence16.INIT_FCS_16

        for i in range(start, start + length):
            index = (fcs ^ data[i]) & 0xFF
            fcs = (fcs >> 8) ^ FastFrameCheckSequence16.fast_frame_check_crc_table[
                index
            ]

        return fcs ^ 0xFFFF  # complement



"""Serial connection factory."""
from asyncio import AbstractEventLoop, get_event_loop
from asyncio import Queue  # pylint: disable=unused-import
from typing import Optional, cast

import serial_asyncio  # type: ignore

from amshan.hdlc import HdlcFrame  # pylint: disable=unused-import
from amshan.meter_connection import (
    MeterTransportProtocol,
    SmartMeterFrameContentProtocol,
    SmartMeterFrameProtocol,
)


async def create_serial_frame_connection(
    queue: "Queue[HdlcFrame]",
    loop: Optional[AbstractEventLoop],
    *args,
    **kwargs,
) -> MeterTransportProtocol:
    """
    Create serial connection using SmartMeterFrameProtocol.

    :param queue: Queue for received frames
    :param loop: The event handler
    :param args: Passed to serial_asyncio.create_serial_connection and further to serial.Serial init function
    :param kwargs: Passed to serial_asyncio.create_serial_connection and further the serial.Serial init function
    :return: Tuple of transport and protocol
    """
    return cast(
        MeterTransportProtocol,
        await serial_asyncio.create_serial_connection(
            loop if loop else get_event_loop(),
            lambda: SmartMeterFrameProtocol(queue),
            *args,
            **kwargs,
        ),
    )


async def create_serial_frame_content_connection(
    queue: "Queue[bytes]",
    loop: Optional[AbstractEventLoop],
    *args,
    **kwargs,
) -> MeterTransportProtocol:
    """Create serial connection using SmartMeterFrameContentProtocol.

    :param queue: Queue for received frames
    :param loop: The event handler
    :param args: Passed to serial_asyncio.create_serial_connection and further to serial.Serial init function
    :param kwargs: Passed to serial_asyncio.create_serial_connection and further the serial.Serial init function
    :return: Tuple of transport and protocol
    """
    return cast(
        MeterTransportProtocol,
        await serial_asyncio.create_serial_connection(
            loop if loop else get_event_loop(),
            lambda: SmartMeterFrameContentProtocol(queue),
            *args,
            **kwargs,
        ),
    )



"""Use this module to read HDLC frames."""
import logging
from typing import List, Optional, cast

from amshan import fastframecheck

_LOGGER = logging.getLogger(__name__)


class HdlcFrameHeader:
    """The start (header) of an HdlcFrame."""

    def __init__(self, frame: "HdlcFrame") -> None:
        """
        Initialize header.

        Used by parent frame.
        """
        self._frame: HdlcFrame = frame
        self._control_position: Optional[int] = None
        self._is_header_good: Optional[bool] = None

    def update(self) -> None:
        """Update fields when more frame data has been read. Used by parent HdlcFrame."""
        if self._control_position is None and len(self._frame) > 3:
            self._control_position = self._get_control_field_position()

        if self._control_position is not None:
            if self._is_header_good is None:
                if len(self._frame) == self._control_position + 3:
                    self._is_header_good = self._frame.is_good_ffc

    @property
    def frame_format(self) -> Optional[int]:
        """Return the value of frame format if the value has been read."""
        # The length of the frame format field is two bytes. It consists of three sub-fields referred to as the Format
        # type sub-field (4 bit), the Segmentation bit (S, 1 bit) and the frame length sub-field (11 bit).
        if len(self._frame) >= 2:
            return self._frame.frame_data[0] << 8 | self._frame.frame_data[1]
        return None

    @property
    def frame_format_type(self) -> Optional[int]:
        """Return the value of frame format type sub-field when frame format has been read."""
        if self.frame_format is not None:
            return (self.frame_format >> 12) & 0b1111
        return None

    @property
    def segmentation(self) -> Optional[bool]:
        """Return the value of frame format Segmentation flag when frame format has been read."""
        if self.frame_format is not None:
            return ((self.frame_format >> 11) & 0x1) == 0x1
        return None

    @property
    def frame_length(self) -> Optional[int]:
        """
        Return the value of frame format length sub-field when frame format has been read.

        The value of the frame length subfield is the count of octets in the frame
        excluding the opening and closing frame flag sequences.
        """
        if self.frame_format is not None:
            return self.frame_format & 0b11111111111
        return None

    @property
    def destination_address(self) -> Optional[bytes]:
        """
        Return the value of destination address when the field has been read.

        Depending on the direction of the data transfer, both the client and the server addresses can
        be destination or source addresses.

        The client address shall always be expressed on one byte.
        """
        if len(self._frame) >= 2:
            return self._get_address(2)
        return None

    @property
    def source_address(self) -> Optional[bytes]:
        """
        Return the value of source address when the field has been read.

        Depending on the direction of the data transfer, both the client and the server addresses can
        be destination or source addresses.

        The client address shall always be expressed on one byte.
        """
        destination_adr = self.destination_address
        if destination_adr is not None:
            return self._get_address(2 + len(destination_adr))
        return None

    @property
    def control(self) -> Optional[int]:
        """
        Return the value of control field when the field has been read.

        It indicates the type of commands or responses,
        and contains sequence numbers, where appropriate (frames I, RR and RNR).
        """
        is_available = (
            self._control_position is not None
            and len(self._frame) > self._control_position
        )
        if is_available:
            return self._frame.frame_data[cast(int, self._control_position)]
        return None

    @property
    def header_check_sequence(self) -> Optional[int]:
        """
        Header check sequence (HCS) field when the field has been read.

        This check sequence is applied to only the header,
        i.e., the bits between the opening flag sequence and the header check sequence.
        """
        is_available = (
            self._control_position is not None
            and len(self._frame) > self._control_position + 2
        )
        if is_available:
            return (
                self._frame.frame_data[cast(int, self._control_position) + 1] << 8
                | self._frame.frame_data[cast(int, self._control_position) + 2]
            )
        return None

    @property
    def information_position(self) -> Optional[int]:
        """
        Information field position when the field position is known and is available.

        Frames that do not have an information field or have an empty information field,
        e.g., as with some supervisory frames, do not contain an HCS and FCS, only an FCS.
        """
        if self._control_position is not None:
            return self._control_position + 3
        return None

    def _get_address(self, position: int) -> Optional[bytes]:
        """Get variable length address from position."""
        # As specified in ISO/IEC 13239:2002, 4.7.1, The address field range can be extended by reserving the first
        # transmitted bit (low-order) of each address octet which would then be set to binary zero to indicate that
        # the following octet is an extension of the address field. The format of the extended octet(s) shall be the
        # same as that of the first octet. Thus, the address field may be recursively extended. The last octet of an
        # address field is indicted by setting the low-order bit to binary one.

        if len(self._frame) > position:
            adr = bytearray()

            i = position
            frame_data = self._frame.frame_data
            while True:
                if i >= len(self._frame):
                    return None

                current = frame_data[i]
                adr.append(current)

                if (current & 0x01) == 0x01:
                    return bytes(adr)

                i += 1

        return None

    def _get_control_field_position(self) -> Optional[int]:
        destination_adr = self.destination_address
        if destination_adr is not None:
            source_adr = self.source_address
            if source_adr is not None:
                return 2 + len(destination_adr) + len(source_adr)
        return None


class HdlcFrame:
    """
    Use this class to read HDLC frames by calling append for each byte.

    is_good is true when checksum is ok. This can be the case both when header is read
    and when done reading information.
    """

    def __init__(self) -> None:
        """Construct HdlcFrame."""
        self._frame_data = bytearray()
        self._ffc = fastframecheck.FastFrameCheckSequence16()
        self._escape_next = False
        self._header = HdlcFrameHeader(self)

    # Frame length is specified with 11 bit.
    MAX_FRAME_LENGTH: int = 0b11111111111

    def __len__(self) -> int:
        """
        Length of currently parsed bytes.

        Note that expected complete frame length is found in header when header has been read.
        """
        return len(self._frame_data)

    def append(self, byte: int) -> None:
        """Append byte to frame."""
        self._frame_data.append(byte)
        self._ffc.update(byte)
        self._header.update()

    @property
    def frame_data(self) -> bytes:
        """
        Return frame data bytes.

        Data has been unescaped when the reader uses octet frame stuffing (see constructor).
        """
        return bytes(self._frame_data)

    @property
    def is_good_ffc(self) -> bool:
        """
        Return Current Fast Frame Check calculation state.

        Can be true for both end of header and body.
        """
        return self._ffc.is_good

    @property
    def is_expected_length(self) -> bool:
        """Return True when current length is the same as length from header."""
        return self._header.frame_length == len(self)

    @property
    def header(self) -> HdlcFrameHeader:
        """Frame header."""
        return self._header

    @property
    def frame_check_sequence(self) -> Optional[int]:
        """Frame check sequence if complete frame has been read. None or invalid number if not."""
        if self._header.information_position is not None:
            if len(self._frame_data) >= self._header.information_position:
                return (
                    self._frame_data[len(self._frame_data) - 2] << 8
                    | self._frame_data[len(self._frame_data) - 1]
                )
        return None

    @property
    def information(self) -> Optional[bytes]:
        """Information field when the field has been read and is available."""
        info_position = self._header.information_position
        if info_position is not None and len(self._frame_data) > info_position:
            return bytes(self._frame_data[info_position:-2])
        return None


class HdlcFrameReader:
    """Use this class to HDLC-frames as stream of bytes."""

    # The Control Escape octet is defined as binary 01111101 (hexadecimal 0x7d)
    CONTROL_ESCAPE: int = 0x7D

    # The Flag Sequence indicates the beginning or end of a frame.
    # The octet stream is examined on an octet-by-octet basis for the value 01111110 (hexadecimal 0x7e).
    FLAG_SEQUENCE: int = 0x7E

    def __init__(
        self, use_octet_stuffing: bool = False, use_abort_sequence: bool = False
    ) -> None:
        """
        Construct HHdlcFrameReader.

        :param use_octet_stuffing: true to use octet stuffing (0x7D as escape octet)
        """
        self._use_octet_stuffing = use_octet_stuffing
        self._use_abort_sequence = use_abort_sequence
        self._unescape_next = False
        self._buffer = _ReaderBuffer()
        self._raw_frame_data = bytearray()
        self._frame: Optional[HdlcFrame] = None

    @property
    def unescape_next(self) -> bool:
        """Return True if octet stuffing is used and next octet is escaped and must be unescaped."""
        return self._unescape_next

    @property
    def is_in_hunt_mode(self) -> bool:
        """Return True when reader is hunting for start of frame."""
        return self._frame is None

    def read(self, data_chunk: bytes) -> List[HdlcFrame]:
        """
        Call this function to read chunks of bytes.

        :param data_chunk: next bytes to parsed.
        :return: frame when a frame is complete (both with correct and incorrect checksum).
        """
        frames_received: List[HdlcFrame] = []

        self._buffer.extend(data_chunk)

        if self._frame is None:  # in hunt mode
            self._buffer.trim_buffer_to_flag_or_end()

        while self._buffer.is_available:
            frame_complete = self._read_next()
            if frame_complete:
                _LOGGER.debug(
                    "Frame of %s length %d received with %s checksum.",
                    "expected"
                    if cast(HdlcFrame, self._frame).is_expected_length
                    else "unexpected",
                    len(cast(HdlcFrame, self._frame)),
                    "good" if cast(HdlcFrame, self._frame).is_good_ffc else "bad",
                )
                frames_received.append(cast(HdlcFrame, self._frame))
                self._start_frame()
                self._buffer.trim_buffer_to_current_position()

        return frames_received

    def _read_next(self) -> bool:
        frame_complete = False

        current = self._buffer.pop()
        is_flag = current == self.FLAG_SEQUENCE
        if is_flag:
            frame_complete = self._handle_flag_sequence()
        elif self._frame is not None:  # not in hunt mode
            self._append_to_frame(current)
            if len(self._frame) > HdlcFrame.MAX_FRAME_LENGTH:
                _LOGGER.debug(
                    "Max frame length reached. Discard frame: %s",
                    self._raw_frame_data.hex(),
                )
                self._goto_hunt_mode()
                frame_complete = False

        return frame_complete

    def _handle_flag_sequence(self) -> bool:
        frame_complete = False

        if self._frame is None:
            _LOGGER.debug("Found flag sequence in frame hunt mode")
            self._start_frame()

        elif len(self._frame) == 0:
            # Found new flag sequence. Two is normal ( end + start), one is allowed, and many possible if time fill.
            pass

        elif self._frame.header.header_check_sequence is None:
            # Frames which are too short are silently discarded, and not counted as a FCS error.
            _LOGGER.debug(
                "Found flag sequence. Too short frame (%d bytes). Discard frame: %s",
                len(self._frame),
                self._raw_frame_data.hex(),
            )
            self._goto_hunt_mode()

        # check if previous octet was Control Escape if abort sequence is activated
        elif (
            self._use_abort_sequence
            and len(self._raw_frame_data) > 1
            and self._raw_frame_data[-1:][0] == self.CONTROL_ESCAPE
        ):
            # Frames which end with a Control Escape octet
            # followed immediately by a closing Flag Sequence,
            # are silently discarded, and not counted as a FCS error.
            _LOGGER.debug(
                "Abort sequence. Discard frame: %s", self._raw_frame_data.hex()
            )
            self._goto_hunt_mode()

        elif self._use_octet_stuffing:
            # Control Escape is never in content when using octet stuffing
            frame_complete = True

        elif self._frame.is_expected_length:
            frame_complete = True

        else:
            self._append_to_frame(self.FLAG_SEQUENCE)

        return frame_complete

    def _append_to_frame(self, current: int) -> None:
        assert self._frame is not None
        self._raw_frame_data.append(current)
        if self._use_octet_stuffing:
            if self._unescape_next:
                self._unescape_next = False
                unescaped = current ^ 0x20
                self._frame.append(unescaped)
            else:
                if current == HdlcFrameReader.CONTROL_ESCAPE:
                    self._unescape_next = True
                else:
                    self._frame.append(current)
        else:
            self._frame.append(current)

    def _start_frame(self) -> None:
        self._frame = HdlcFrame()
        self._raw_frame_data.clear()

    def _goto_hunt_mode(self) -> None:
        self._frame = None
        self._buffer.trim_buffer_to_flag_or_end()


class _ReaderBuffer:
    """Buffer class used by HdlcFrameReader."""

    def __init__(self) -> None:
        self._buffer = bytearray()
        self._buffer_pos = 0

    @property
    def is_available(self) -> bool:
        """Byte(s) are available."""
        return len(self._buffer) > self._buffer_pos

    def pop(self) -> int:
        """Pop one byte from buffer."""
        byte = self._buffer[self._buffer_pos]
        self._buffer_pos += 1
        return byte

    def extend(self, data_chunk: bytes) -> None:
        """Add bytes to buffer."""
        self._buffer.extend(data_chunk)

    def trim_buffer_to_current_position(self) -> None:
        """Trim buffer to current position."""
        self._buffer = self._buffer[self._buffer_pos :]
        self._buffer_pos = 0

    def trim_buffer_to_flag_or_end(self) -> None:
        """Trim buffer to flag sequence or end of buffer."""
        self.trim_buffer_to_current_position()
        flag_pos = self._buffer.find(HdlcFrameReader.FLAG_SEQUENCE)
        if flag_pos == -1:
            # flag sequence not found
            self._buffer.clear()
        if flag_pos > 0:
            # trim data before flag sequence
            self._buffer = self._buffer[flag_pos:]
        self._buffer_pos = 0



"""Decoding support for Kaifa meters."""
# pylint: disable=protected-access
from datetime import datetime
from typing import Dict, List, Union

import construct  # type: ignore

from amshan import cosem, obis_map

NotificationBody: construct.Struct = construct.Struct(
    construct.Const(
        cosem.CommonDataTypes.structure, cosem.CommonDataTypes
    ),  # expect structure
    "length" / construct.Int8ub,
    "list_items"
    / construct.Array(
        construct.this.length,
        construct.Struct(
            "index" / construct.Computed(construct.this._index),
            "_value_type" / cosem.CommonDataTypes,
            "value"
            / construct.Switch(
                construct.this._value_type,
                {
                    # octed string is used for both text and date time
                    cosem.CommonDataTypes.octet_string: construct.IfThenElse(
                        construct.this._index < 4, cosem.OctedStringText, cosem.DateTime
                    ),
                    cosem.CommonDataTypes.double_long_unsigned: cosem.DoubleLongUnsigned,
                },
            ),
        ),
    ),
)

LlcPdu: construct.Struct = cosem.get_llc_pdu_struct(NotificationBody)


def _get_field_lists() -> List[List[str]]:
    item_order_list_3_three_phase = [
        obis_map.NEK_HAN_FIELD_OBIS_LIST_VER_ID,
        obis_map.NEK_HAN_FIELD_METER_ID,
        obis_map.NEK_HAN_FIELD_METER_TYPE,
        obis_map.NEK_HAN_FIELD_ACTIVE_POWER_IMPORT,
        obis_map.NEK_HAN_FIELD_ACTIVE_POWER_EXPORT,
        obis_map.NEK_HAN_FIELD_REACTIVE_POWER_IMPORT,
        obis_map.NEK_HAN_FIELD_REACTIVE_POWER_EXPORT,
        obis_map.NEK_HAN_FIELD_CURRENT_L1,
        obis_map.NEK_HAN_FIELD_CURRENT_L2,
        obis_map.NEK_HAN_FIELD_CURRENT_L3,
        obis_map.NEK_HAN_FIELD_VOLTAGE_L1,
        obis_map.NEK_HAN_FIELD_VOLTAGE_L2,
        obis_map.NEK_HAN_FIELD_VOLTAGE_L3,
        obis_map.NEK_HAN_FIELD_METER_DATETIME,
        obis_map.NEK_HAN_FIELD_ACTIVE_POWER_IMPORT_HOUR,
        obis_map.NEK_HAN_FIELD_ACTIVE_POWER_EXPORT_HOUR,
        obis_map.NEK_HAN_FIELD_REACTIVE_POWER_IMPORT_HOUR,
        obis_map.NEK_HAN_FIELD_REACTIVE_POWER_EXPORT_HOUR,
    ]

    item_order_list_3_single_phase = (
        item_order_list_3_three_phase[:8]
        + item_order_list_3_three_phase[10:11]
        + item_order_list_3_three_phase[13:]
    )

    item_order_list_2_single_phase = item_order_list_3_single_phase[:-5]

    item_order_list_2_three_phase = item_order_list_3_three_phase[:-5]

    return [
        [obis_map.NEK_HAN_FIELD_ACTIVE_POWER_IMPORT],
        item_order_list_2_single_phase,
        item_order_list_2_three_phase,
        item_order_list_3_single_phase,
        item_order_list_3_three_phase,
    ]


_field_order_lists: List[List[str]] = _get_field_lists()

_FIELD_SCALING = {
    obis_map.NEK_HAN_FIELD_CURRENT_L1: -3,
    obis_map.NEK_HAN_FIELD_CURRENT_L2: -3,
    obis_map.NEK_HAN_FIELD_CURRENT_L3: -3,
    obis_map.NEK_HAN_FIELD_VOLTAGE_L1: -1,
    obis_map.NEK_HAN_FIELD_VOLTAGE_L2: -1,
    obis_map.NEK_HAN_FIELD_VOLTAGE_L3: -1,
}


def normalize_parsed_frame(
    frame: construct.Struct,
) -> Dict[str, Union[str, int, float, datetime]]:
    """Convert data from meters construct structure to a dictionary with common key names."""
    list_items = frame.information.notification_body.list_items

    current_list_names: List[str] = next(
        (x for x in _field_order_lists if len(x) == len(list_items)), []
    )

    dictionary = {
        obis_map.NEK_HAN_FIELD_METER_MANUFACTURER: "Kaifa",
        obis_map.NEK_HAN_FIELD_METER_DATETIME: frame.information.DateTime.datetime,
    }

    for measure in list_items:
        element_name = current_list_names[measure.index]

        if element_name == obis_map.NEK_HAN_FIELD_METER_DATETIME:
            dictionary[element_name] = measure.value.datetime
        else:
            scale = _FIELD_SCALING.get(element_name, None)
            if scale:
                scaled_value = measure.value * (10 ** scale)
                dictionary[element_name] = scaled_value
            else:
                dictionary[element_name] = measure.value

    return dictionary


def decode_frame_content(
    frame_content: bytes,
) -> Dict[str, Union[str, int, float, datetime]]:
    """Decode meter LLC PDU frame content as a dictionary."""
    parsed = LlcPdu.parse(frame_content)
    return normalize_parsed_frame(parsed)



"""Decoding support for Kamstrup meters."""
# pylint: disable=protected-access
from datetime import datetime
from typing import Dict, Union

import construct  # type: ignore

from amshan import cosem, obis_map

Element: construct.Struct = construct.Struct(
    "_element_type" / construct.Peek(cosem.CommonDataTypes),
    "obis"
    / construct.If(
        cosem.CommonDataTypes.octet_string == construct.this._element_type,
        cosem.ObisCodeOctedStringField,
    ),
    "value_type" / construct.Peek(cosem.CommonDataTypes),
    "value"
    / construct.IfThenElse(
        cosem.CommonDataTypes.octet_string == construct.this.value_type,
        cosem.DateTimeField,
        cosem.Field,
    ),
)

NotificationBody: construct.Struct = construct.Struct(
    construct.Const(
        cosem.CommonDataTypes.structure, cosem.CommonDataTypes
    ),  # expect structure
    "length" / construct.Int8ub,
    "list_items" / construct.GreedyRange(Element),
)

LlcPdu: construct.Struct = cosem.get_llc_pdu_struct(NotificationBody)

_field_scaling = {
    "1.1.31.7.0.255": -2,
    "1.1.51.7.0.255": -2,
    "1.1.71.7.0.255": -2,
}


def normalize_parsed_frame(
    frame: construct.Struct,
) -> Dict[str, Union[str, int, float, datetime]]:
    """Convert data from meters construct structure to a dictionary with common key names."""
    dictionary = {
        obis_map.NEK_HAN_FIELD_METER_MANUFACTURER: "Kamstrup",
        obis_map.NEK_HAN_FIELD_METER_DATETIME: frame.information.DateTime.datetime,
    }

    list_items = frame.information.notification_body.list_items
    for measure in list_items:
        # list version is the only element without obis code
        element_name = (
            obis_map.obis_name_map[measure.obis]
            if measure.obis
            else obis_map.NEK_HAN_FIELD_OBIS_LIST_VER_ID
        )

        if element_name == obis_map.NEK_HAN_FIELD_METER_DATETIME:
            dictionary[element_name] = measure.value.datetime
        else:
            if isinstance(measure.value, int):
                scale = _field_scaling.get(measure.obis, None)
                if scale:
                    dictionary[element_name] = measure.value * (10 ** scale)
                else:
                    dictionary[element_name] = measure.value
            else:
                dictionary[element_name] = measure.value

    return dictionary


def decode_frame_content(
    frame_content: bytes,
) -> Dict[str, Union[str, int, float, datetime]]:
    """Decode meter LLC PDU frame content as a dictionary."""
    parsed = LlcPdu.parse(frame_content)
    return normalize_parsed_frame(parsed)



"""Python package setup."""
import setuptools

with open("README.md", "r", encoding="UTF-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="amshan",
    version="1.0.0",
    author="Tore Amundsen",
    author_email="tore@amundsen.org",
    description="Decode MBUS (Meter Bus) data with special support for norwegian AMS-smart meters (HAN port)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    platforms=["POSIX", "Windows"],
    url="https://github.com/toreamun/amshan",
    packages=["amshan"],
    package_data={"amshan": ["py.typed"]},
    keywords=[
        "meter",
        "han",
        "ams",
        "mbus",
        "aidon",
        "kaifa",
        "kamstrup",
        "cosem",
        "hdlc",
        "fast frame check",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=["construct"],
    extras_require={"serial": ["pyserial-asyncio>=0.4"]},
)
